from logging import LoggerAdapter
from typing import Iterable

import aiohttp_cors
import jwt
from aiohttp import web
from pydantic import BaseModel

from ..defs import RootContext
from ..types import CORSOptions, PydanticResponse, WebMiddleware
from ..utils import ensure_json_serializable
from .types import ConfRequestModel
from .utils import pydantic_api_handler


class TokenResponseModel(BaseModel):
    token: str


@pydantic_api_handler(ConfRequestModel)
async def conf_v2(
    request: web.Request, params: ConfRequestModel
) -> PydanticResponse[TokenResponseModel]:
    """
    Generates and returns a token which will be used as an authentication credential for
     /v2/proxy/{token}/{session}/add request.
    """
    log: LoggerAdapter = request["log"]

    root_ctx: RootContext = request.app["_root.context"]

    assert params.session.id and params.session.access_key, "Not meant for inference apps"

    token = jwt.encode(
        ensure_json_serializable({
            "login_session_token": params.login_session_token,
            "kernel_host": params.kernel_host,
            "kernel_port": params.kernel_port,
            "session_id": params.session.id,
            "user_uuid": params.session.user_uuid,
            "group_id": params.session.group_id,
            "access_key": params.session.access_key,
            "domain_name": params.session.domain_name,
        }),
        root_ctx.local_config.wsproxy.jwt_encrypt_key,
    )
    log.debug("built token with body {}", params.model_dump())

    return PydanticResponse(TokenResponseModel(token=token))


async def init(app: web.Application) -> None:
    pass


async def shutdown(app: web.Application) -> None:
    pass


def create_app(
    default_cors_options: CORSOptions,
) -> tuple[web.Application, Iterable[WebMiddleware]]:
    app = web.Application()
    app["prefix"] = "v2/conf"
    app.on_startup.append(init)
    app.on_shutdown.append(shutdown)
    cors = aiohttp_cors.setup(app, defaults=default_cors_options)
    root_resource = cors.add(app.router.add_resource(r""))
    cors.add(root_resource.add_route("POST", conf_v2))
    return app, []



import asyncio
import functools
import grp
import importlib
import importlib.resources
import logging
import os
import pwd
import signal
import sys
import traceback
import uuid
from contextlib import asynccontextmanager as actxmgr
from logging import LoggerAdapter
from pathlib import Path
from typing import Any, AsyncGenerator, AsyncIterator, Final, Iterable, Mapping, Sequence, cast

import aiohttp_cors
import aiohttp_jinja2
import aiomonitor
import aiotools
import click
import jinja2
from aiohttp import web
from aiohttp.typedefs import Middleware
from setproctitle import setproctitle

from ai.backend.common.metrics.http import (
    build_api_metric_middleware,
    build_prometheus_metrics_handler,
)
from ai.backend.common.metrics.metric import CommonMetricRegistry
from ai.backend.common.metrics.profiler import Profiler, PyroscopeArgs
from ai.backend.common.msgpack import DEFAULT_PACK_OPTS, DEFAULT_UNPACK_OPTS
from ai.backend.common.utils import env_info
from ai.backend.logging import BraceStyleAdapter, Logger, LogLevel
from ai.backend.wsproxy.exceptions import (
    BackendError,
    GenericBadRequest,
    InternalServerError,
    MethodNotAllowed,
    URLNotFound,
)
from ai.backend.wsproxy.types import (
    AppCreator,
    ProxyProtocol,
    WebRequestHandler,
)

from . import __version__
from .config import ServerConfig
from .config import load as load_config
from .defs import CleanupContext, RootContext
from .proxy.frontend import (
    HTTPPortFrontend,
    TCPFrontend,
)
from .utils import (
    config_key_to_kebab_case,
    ensure_json_serializable,
    mime_match,
)

log = BraceStyleAdapter(logging.getLogger(__spec__.name))

REDIS_APPPROXY_DB: Final[int] = 10  # FIXME: move to ai.backend.common.defs
EVENT_DISPATCHER_CONSUMER_GROUP: Final[str] = "appwsproxy"

global_subapp_pkgs: Final[list[str]] = [
    ".circuit",
    ".conf",
    ".endpoint",
    ".proxy",
    ".setup",
]


@web.middleware
async def request_context_aware_middleware(
    request: web.Request, handler: WebRequestHandler
) -> web.StreamResponse:
    request_id = request.headers.get("X-BackendAI-RequestID", str(uuid.uuid4()))
    request["request_id"] = request_id
    request["log"] = BraceStyleAdapter(logging.getLogger(f"{__spec__.name} - #{request_id}"))  # type: ignore[name-defined]
    resp = await handler(request)
    return resp


@web.middleware
async def api_middleware(request: web.Request, handler: WebRequestHandler) -> web.StreamResponse:
    _handler = handler
    method_override = request.headers.get("X-Method-Override", None)
    if method_override:
        request = request.clone(method=method_override)
        new_match_info = await request.app.router.resolve(request)
        if new_match_info is None:
            raise InternalServerError("No matching method handler found")
        _handler = new_match_info.handler
        request._match_info = new_match_info  # type: ignore  # this is a hack
    ex = request.match_info.http_exception
    if ex is not None:
        # handled by exception_middleware
        raise ex
    request_id = request.headers.get("X-BackendAI-RequestID", str(uuid.uuid4()))
    request["request_id"] = request_id
    request["log"] = BraceStyleAdapter(logging.getLogger(f"{__spec__.name} - #{request_id}"))  # type: ignore[name-defined]
    resp = await _handler(request)
    return resp


@web.middleware
async def exception_middleware(
    request: web.Request, handler: WebRequestHandler
) -> web.StreamResponse:
    root_ctx: RootContext = request.app["_root.context"]
    log: LoggerAdapter = request["log"]

    try:
        resp = await handler(request)
    except BackendError as ex:
        if ex.status_code == 500:
            log.exception("Internal server error raised inside handlers")
        if mime_match(request.headers.get("accept", "text/html"), "application/json", strict=True):
            return web.json_response(
                ensure_json_serializable(ex.body_dict),
                status=ex.status_code,
            )
        else:
            return aiohttp_jinja2.render_template(
                "error.jinja2",
                request,
                ex.body_dict,
                status=ex.status_code,
            )
    except web.HTTPException as ex:
        if ex.status_code == 404:
            raise URLNotFound(extra_data=request.path)
        if ex.status_code == 405:
            concrete_ex = cast(web.HTTPMethodNotAllowed, ex)
            raise MethodNotAllowed(
                method=concrete_ex.method, allowed_methods=concrete_ex.allowed_methods
            )
        log.warning("Bad request: {0!r}", ex)
        raise GenericBadRequest
    except asyncio.CancelledError as e:
        # The server is closing or the client has disconnected in the middle of
        # request.  Atomic requests are still executed to their ends.
        log.debug("Request cancelled ({0} {1})", request.method, request.rel_url)
        raise e
    except Exception as e:
        log.exception("Uncaught exception in HTTP request handlers {0!r}", e)
        if root_ctx.local_config.debug.enabled:
            raise InternalServerError(traceback.format_exc())
        else:
            raise InternalServerError()
    else:
        return resp


@actxmgr
async def proxy_frontend_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    match root_ctx.local_config.wsproxy.protocol:
        case ProxyProtocol.HTTP:
            root_ctx.proxy_frontend = HTTPPortFrontend(root_ctx)
        case ProxyProtocol.TCP:
            root_ctx.proxy_frontend = TCPFrontend(root_ctx)
        case _:
            log.error("Unsupported protocol {}", root_ctx.local_config.wsproxy.protocol)
    await root_ctx.proxy_frontend.start()
    log.debug("started proxy protocol {}", root_ctx.proxy_frontend.__class__.__name__)
    yield
    await root_ctx.proxy_frontend.terminate_all_circuits()
    await root_ctx.proxy_frontend.stop()


async def hello(request: web.Request) -> web.Response:
    """
    Returns the API version number.
    """
    return web.json_response({
        "wsproxy": __version__,
    })


async def status(request: web.Request) -> web.Response:
    request["do_not_print_access_log"] = True
    root_ctx: RootContext = request.app["_root.context"]

    config = root_ctx.local_config.wsproxy
    return web.json_response({
        "api_version": "v2",
        "advertise_address": config.advertised_api_addr,
    })


async def on_prepare(request: web.Request, response: web.StreamResponse) -> None:
    response.headers["Server"] = "BackendAI"


def handle_loop_error(
    root_ctx: RootContext,
    loop: asyncio.AbstractEventLoop,
    context: Mapping[str, Any],
) -> None:
    exception = context.get("exception")
    msg = context.get("message", "(empty message)")
    if exception is not None:
        if sys.exc_info()[0] is not None:
            log.exception("Error inside event loop: {0}", msg)
        else:
            exc_info = (type(exception), exception, exception.__traceback__)
            log.error("Error inside event loop: {0}", msg, exc_info=exc_info)


def _init_subapp(
    pkg_name: str,
    root_app: web.Application,
    subapp: web.Application,
    global_middlewares: Iterable[Middleware],
) -> None:
    subapp.on_response_prepare.append(on_prepare)

    async def _set_root_ctx(subapp: web.Application):
        # Allow subapp's access to the root app properties.
        # These are the public APIs exposed to plugins as well.
        subapp["_root.context"] = root_app["_root.context"]

    # We must copy the public interface prior to all user-defined startup signal handlers.
    subapp.on_startup.insert(0, _set_root_ctx)
    if "prefix" not in subapp:
        subapp["prefix"] = pkg_name.split(".")[-1].replace("_", "-")
    prefix = subapp["prefix"]
    root_app.add_subapp("/" + prefix, subapp)
    root_app.middlewares.extend(global_middlewares)


def init_subapp(pkg_name: str, root_app: web.Application, create_subapp: AppCreator) -> None:
    root_ctx: RootContext = root_app["_root.context"]
    subapp, global_middlewares = create_subapp(root_ctx.cors_options)
    _init_subapp(pkg_name, root_app, subapp, global_middlewares)


def build_root_app(
    pidx: int,
    local_config: ServerConfig,
    *,
    cleanup_contexts: Sequence[CleanupContext] | None = None,
    subapp_pkgs: Sequence[str] = [],
) -> web.Application:
    Profiler(
        pyroscope_args=PyroscopeArgs(
            enabled=local_config.pyroscope.enabled,
            application_name=local_config.pyroscope.app_name,
            server_address=local_config.pyroscope.server_addr,
            sample_rate=local_config.pyroscope.sample_rate,
        )
    )
    metric_registry = CommonMetricRegistry.instance()
    app = web.Application(
        middlewares=[
            request_context_aware_middleware,
            exception_middleware,
            api_middleware,
            build_api_metric_middleware(metric_registry.api),
        ]
    )
    root_ctx = RootContext()
    global_exception_handler = functools.partial(handle_loop_error, root_ctx)
    loop = asyncio.get_running_loop()
    loop.set_exception_handler(global_exception_handler)
    app["_root.context"] = root_ctx
    root_ctx.local_config = local_config
    root_ctx.pidx = pidx
    root_ctx.cors_options = {
        "*": aiohttp_cors.ResourceOptions(
            allow_credentials=False, expose_headers="*", allow_headers="*"
        ),
    }
    app.on_response_prepare.append(on_prepare)

    with importlib.resources.as_file(importlib.resources.files("ai.backend.wsproxy")) as f:
        template_path = f / "templates"
        aiohttp_jinja2.setup(app, loader=jinja2.FileSystemLoader(template_path))

    if cleanup_contexts is None:
        cleanup_contexts = [
            proxy_frontend_ctx,
        ]

    async def _cleanup_context_wrapper(cctx, app: web.Application) -> AsyncIterator[None]:
        # aiohttp's cleanup contexts are just async generators, not async context managers.
        cctx_instance = cctx(app["_root.context"])
        app["_cctx_instances"].append(cctx_instance)
        try:
            async with cctx_instance:
                yield
        except Exception as e:
            exc_info = (type(e), e, e.__traceback__)
            log.error("Error initializing cleanup_contexts: {0}", cctx.__name__, exc_info=exc_info)

    async def _call_cleanup_context_shutdown_handlers(app: web.Application) -> None:
        for cctx in app["_cctx_instances"]:
            if hasattr(cctx, "shutdown"):
                try:
                    await cctx.shutdown()
                except Exception:
                    log.exception("error while shutting down a cleanup context")

    app["_cctx_instances"] = []
    app.on_shutdown.append(_call_cleanup_context_shutdown_handlers)
    for cleanup_ctx in cleanup_contexts:
        app.cleanup_ctx.append(
            functools.partial(_cleanup_context_wrapper, cleanup_ctx),
        )
    cors = aiohttp_cors.setup(app, defaults=root_ctx.cors_options)
    # should be done in create_app() in other modules.
    cors.add(app.router.add_route("GET", r"", hello))
    cors.add(app.router.add_route("GET", r"/", hello))
    cors.add(app.router.add_route("GET", "/status", status))
    if subapp_pkgs is None:
        subapp_pkgs = []
    for pkg_name in subapp_pkgs:
        if pidx == 0:
            log.info("Loading module: {0}", pkg_name[1:])
        subapp_mod = importlib.import_module(pkg_name, "ai.backend.wsproxy.api")
        init_subapp(pkg_name, app, getattr(subapp_mod, "create_app"))
    return app


def build_internal_app() -> web.Application:
    app = web.Application()
    metric_registry = CommonMetricRegistry.instance()
    app.router.add_route("GET", "/metrics", build_prometheus_metrics_handler(metric_registry))
    return app


@actxmgr
async def server_main(
    loop: asyncio.AbstractEventLoop,
    pidx: int,
    _args: tuple[ServerConfig, str],
) -> AsyncIterator[None]:
    root_app = build_root_app(pidx, _args[0], subapp_pkgs=global_subapp_pkgs)
    internal_app = build_internal_app()
    root_ctx: RootContext = root_app["_root.context"]

    # Start aiomonitor.
    # Port is set by config (default=50100 + pidx).
    loop.set_debug(root_ctx.local_config.debug.asyncio)
    m = aiomonitor.Monitor(
        loop,
        termui_port=root_ctx.local_config.wsproxy.aiomonitor_termui_port + pidx,
        webui_port=root_ctx.local_config.wsproxy.aiomonitor_webui_port + pidx,
        console_enabled=False,
        hook_task_factory=root_ctx.local_config.debug.enhanced_aiomonitor_task_info,
    )
    m.prompt = f"monitor (wsproxy[{pidx}@{os.getpid()}]) >>> "
    # Add some useful console_locals for ease of debugging
    m.console_locals["root_app"] = root_app
    m.console_locals["root_ctx"] = root_ctx
    aiomon_started = False
    try:
        m.start()
        aiomon_started = True
    except Exception as e:
        log.warning("aiomonitor could not start but skipping this error to continue", exc_info=e)

    # Plugin webapps should be loaded before runner.setup(),
    # which freezes on_startup event.
    try:
        runner = web.AppRunner(root_app, keepalive_timeout=30.0)
        internal_runner = web.AppRunner(internal_app, keepalive_timeout=30.0)
        await runner.setup()
        await internal_runner.setup()
        site = web.TCPSite(
            runner,
            str(root_ctx.local_config.wsproxy.bind_host),
            root_ctx.local_config.wsproxy.bind_api_port,
            backlog=1024,
            reuse_port=True,
        )
        internal_site = web.TCPSite(
            internal_runner,
            str(root_ctx.local_config.wsproxy.bind_host),
            root_ctx.local_config.wsproxy.internal_api_port,
            backlog=1024,
            reuse_port=True,
        )
        await site.start()
        await internal_site.start()

        if os.geteuid() == 0:
            uid = root_ctx.local_config.wsproxy.user
            gid = root_ctx.local_config.wsproxy.group
            os.setgroups([
                g.gr_gid for g in grp.getgrall() if pwd.getpwuid(uid).pw_name in g.gr_mem
            ])
            os.setgid(gid)
            os.setuid(uid)
            log.info("changed process uid and gid to {}:{}", uid, gid)
        log.info(
            "started handling API requests at {}:{}",
            root_ctx.local_config.wsproxy.bind_host,
            root_ctx.local_config.wsproxy.bind_api_port,
        )

        try:
            yield
        finally:
            log.info("shutting down...")
            await runner.cleanup()
    finally:
        if aiomon_started:
            m.close()


@aiotools.server_context
async def server_main_logwrapper(
    loop: asyncio.AbstractEventLoop,
    pidx: int,
    _args: Any,
) -> AsyncGenerator[None, signal.Signals]:
    setproctitle(f"backend.ai: wsproxy worker-{pidx}")
    log_endpoint = _args[1]
    logging_config = config_key_to_kebab_case(_args[0].logging.model_dump(exclude_none=True))
    logging_config["endpoint"] = log_endpoint
    logger = Logger(
        logging_config,
        is_master=False,
        log_endpoint=log_endpoint,
        msgpack_options={
            "pack_opts": DEFAULT_PACK_OPTS,
            "unpack_opts": DEFAULT_UNPACK_OPTS,
        },
    )
    try:
        with logger:
            async with server_main(loop, pidx, _args):
                yield
    except Exception:
        traceback.print_exc()


@click.group(invoke_without_command=True)
@click.option(
    "-f",
    "--config-path",
    "--config",
    type=Path,
    default=None,
    help=("The config file path. (default: ./wsproxy.toml and /etc/backend.ai/wsproxy.toml)"),
)
@click.option(
    "--debug",
    is_flag=True,
    help="This option will soon change to --log-level TEXT option.",
)
@click.option(
    "--log-level",
    type=click.Choice([*LogLevel], case_sensitive=False),
    default=LogLevel.NOTSET,
    help="Set the logging verbosity level",
)
@click.pass_context
def main(
    ctx: click.Context,
    config_path: Path,
    log_level: str,
    debug: bool = False,
) -> None:
    """
    Start the wsproxy service as a foreground process.
    """
    cfg = load_config(config_path, LogLevel.DEBUG if debug else LogLevel[log_level])

    if ctx.invoked_subcommand is None:
        cfg.wsproxy.pid_file.touch(exist_ok=True)
        cfg.wsproxy.pid_file.write_text(str(os.getpid()))
        ipc_base_path = cfg.wsproxy.ipc_base_path
        ipc_base_path.mkdir(exist_ok=True, parents=True)
        log_sockpath = ipc_base_path / f"worker-logger-{os.getpid()}.sock"
        log_endpoint = f"ipc://{log_sockpath}"
        logging_config = config_key_to_kebab_case(cfg.logging.model_dump(exclude_none=True))
        logging_config["endpoint"] = log_endpoint
        try:
            logger = Logger(
                logging_config,
                is_master=True,
                log_endpoint=log_endpoint,
                msgpack_options={
                    "pack_opts": DEFAULT_PACK_OPTS,
                    "unpack_opts": DEFAULT_UNPACK_OPTS,
                },
            )
            with logger:
                setproctitle("backend.ai: wsproxy")
                log.info("Backend.AI WSProxy {0}", __version__)
                log.info("runtime: {0}", env_info())
                log_config = logging.getLogger("ai.backend.wsproxy.config")
                log_config.debug("debug mode enabled.")
                if cfg.wsproxy.event_loop == "uvloop":
                    import uvloop

                    uvloop.install()
                    log.info("Using uvloop as the event loop backend")
                try:
                    aiotools.start_server(
                        server_main_logwrapper,
                        num_workers=1,
                        args=(cfg, log_endpoint),
                        wait_timeout=5.0,
                    )
                finally:
                    log.info("terminated.")
        finally:
            if cfg.wsproxy.pid_file.is_file():
                # check is_file() to prevent deleting /dev/null!
                cfg.wsproxy.pid_file.unlink()
    else:
        # Click is going to invoke a subcommand.
        pass


if __name__ == "__main__":
    sys.exit(main())



import logging
from abc import ABCMeta, abstractmethod
from typing import Generic, TypeVar
from uuid import UUID

from ai.backend.logging import BraceStyleAdapter
from ai.backend.wsproxy.defs import RootContext
from ai.backend.wsproxy.exceptions import ObjectNotFound
from ai.backend.wsproxy.types import (
    Circuit,
    RouteInfo,
    TCircuitKey,
)

from ..backend.abc import AbstractBackend

log = BraceStyleAdapter(logging.getLogger(__spec__.name))

TBackend = TypeVar("TBackend", bound=AbstractBackend)


class AbstractFrontend(Generic[TBackend, TCircuitKey], metaclass=ABCMeta):
    root_context: RootContext
    circuits: dict[TCircuitKey, Circuit]
    backends: dict[TCircuitKey, TBackend]

    def __init__(self, root_context: RootContext) -> None:
        self.root_context = root_context
        self.circuits = {}
        self.backends = {}

    def get_circuit_by_endpoint_id(self, endpoint_id: UUID) -> Circuit:
        for _, circuit in self.circuits.items():
            if circuit.endpoint_id == endpoint_id:
                return circuit
        raise ObjectNotFound(object_name="Circuit")

    def get_circuit_by_id(self, id: UUID) -> Circuit:
        for _, circuit in self.circuits.items():
            if circuit.id == id:
                return circuit
        raise ObjectNotFound(object_name="Circuit")

    async def register_circuit(self, circuit: Circuit, routes: list[RouteInfo]) -> None:
        key = self.get_circuit_key(circuit)
        self.circuits[key] = circuit
        self.backends[key] = await self.initialize_backend(circuit, routes)
        log.info(
            "circuit {} (app:{}, mode: {}) registered", circuit.id, circuit.app, circuit.app_mode
        )

    async def update_circuit_route_info(
        self, circuit: Circuit, new_routes: list[RouteInfo]
    ) -> None:
        key = self.get_circuit_key(circuit)
        assert key in self.circuits, "Slot not active"
        await self.update_backend(self.backends[key], new_routes)

    async def break_circuit(self, circuit: Circuit) -> None:
        key = self.get_circuit_key(circuit)
        assert key in self.circuits, "Slot not active"
        await self.terminate_backend(self.backends[key])
        del self.backends[key]
        del self.circuits[key]
        log.info(
            "circuit {} (app:{}, mode: {}) unregistered", circuit.id, circuit.app, circuit.app_mode
        )

    async def terminate_all_circuits(self) -> None:
        for circuit in list(self.circuits.values()):
            await self.break_circuit(circuit)

    @abstractmethod
    async def start(self) -> None:
        raise NotImplementedError

    @abstractmethod
    async def stop(self) -> None:
        raise NotImplementedError

    @abstractmethod
    async def initialize_backend(self, circuit: Circuit, routes: list[RouteInfo]) -> TBackend:
        raise NotImplementedError

    @abstractmethod
    async def update_backend(self, backend: TBackend, routes: list[RouteInfo]) -> TBackend:
        raise NotImplementedError

    @abstractmethod
    async def terminate_backend(self, backend: TBackend) -> None:
        raise NotImplementedError

    @abstractmethod
    def get_circuit_key(self, circuit: Circuit) -> TCircuitKey:
        raise NotImplementedError



import os
import pwd
import socket
import sys
import types
import typing
from dataclasses import dataclass
from pathlib import Path
from pprint import pformat
from typing import Annotated, Any, Optional

import click
from pydantic import (
    Field,
    GetCoreSchemaHandler,
    GetJsonSchemaHandler,
    ValidationError,
)
from pydantic.json_schema import JsonSchemaValue
from pydantic_core import PydanticUndefined, core_schema

from ai.backend.common import config
from ai.backend.common.config import BaseConfigModel
from ai.backend.logging import LogLevel
from ai.backend.logging.config_pydantic import LoggingConfig

from .types import EventLoopType, ProxyProtocol

_file_perm = (Path(__file__).parent / "server.py").stat()


@dataclass
class UserID:
    default_uid: Optional[int] = None

    @classmethod
    def uid_validator(
        cls,
        value: Optional[int | str],
    ) -> int:
        if value is None:
            assert cls.default_uid, "value is None but default_uid not provided"
            return cls.default_uid
        assert isinstance(value, (int, str)), "value must be an integer"
        match value:
            case int():
                if value == -1:
                    return os.getuid()
                else:
                    return value
            case str():
                try:
                    _value = int(value)
                    if _value == -1:
                        return os.getuid()
                    else:
                        return _value
                except ValueError:
                    try:
                        return pwd.getpwnam(value).pw_uid
                    except KeyError:
                        assert False, f"no such user {value} in system"

    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        _source_type: Any,
        _handler: GetCoreSchemaHandler,
    ) -> core_schema.CoreSchema:
        schema = core_schema.chain_schema([
            core_schema.union_schema([
                core_schema.int_schema(),
                core_schema.str_schema(),
            ]),
            core_schema.no_info_plain_validator_function(cls.uid_validator),
        ])

        return core_schema.json_or_python_schema(
            json_schema=schema,
            python_schema=core_schema.union_schema([
                # check if it's an instance first before doing any further work
                core_schema.union_schema([
                    core_schema.is_instance_schema(int),
                    core_schema.is_instance_schema(str),
                ]),
                schema,
            ]),
            serialization=core_schema.plain_serializer_function_ser_schema(int),
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls, _core_schema: core_schema.CoreSchema, handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        # Use the same schema that would be used for `int`
        return handler(
            core_schema.union_schema([
                core_schema.int_schema(),
                core_schema.str_schema(),
            ])
        )


@dataclass
class GroupID:
    default_gid: int | None = None

    @classmethod
    def uid_validator(
        cls,
        value: int | str | None,
    ) -> int:
        if value is None:
            assert cls.default_gid, "value is None but default_gid not provided"
        assert isinstance(value, (int, str)), "value must be an integer"
        match value:
            case int():
                if value == -1:
                    return os.getgid()
                else:
                    return value
            case str():
                try:
                    _value = int(value)
                    if _value == -1:
                        return os.getgid()
                    else:
                        return _value
                except ValueError:
                    try:
                        return pwd.getpwnam(value).pw_gid
                    except KeyError:
                        assert False, f"no such user {value} in system"

    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        _source_type: Any,
        _handler: GetCoreSchemaHandler,
    ) -> core_schema.CoreSchema:
        schema = core_schema.chain_schema([
            core_schema.union_schema([
                core_schema.int_schema(),
                core_schema.str_schema(),
            ]),
            core_schema.no_info_plain_validator_function(cls.uid_validator),
        ])

        return core_schema.json_or_python_schema(
            json_schema=schema,
            python_schema=core_schema.union_schema([
                # check if it's an instance first before doing any further work
                core_schema.union_schema([
                    core_schema.is_instance_schema(int),
                    core_schema.is_instance_schema(str),
                ]),
                schema,
            ]),
            serialization=core_schema.plain_serializer_function_ser_schema(int),
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls, _core_schema: core_schema.CoreSchema, handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        # Use the same schema that would be used for `int`
        return handler(
            core_schema.union_schema([
                core_schema.int_schema(),
                core_schema.str_schema(),
            ])
        )


class DebugConfig(BaseConfigModel):
    enabled: bool = Field(default=False)
    asyncio: bool = Field(default=False)
    enhanced_aiomonitor_task_info: bool = Field(default=False)
    log_events: bool = Field(default=False)


class WSProxyConfig(BaseConfigModel):
    ipc_base_path: Path = Field(
        default=Path("/tmp/backend.ai/ipc"),
        description="Directory to store temporary UNIX sockets.",
    )
    event_loop: EventLoopType = Field(
        default=EventLoopType.ASYNCIO,
        description="Type of event loop to use.",
    )
    pid_file: Path = Field(
        default=Path(os.devnull),
        description="Place to store process PID.",
        examples=["/run/backend.ai/wsproxy/wsproxy.pid"],
    )

    id: str = Field(
        default=f"i-{socket.gethostname()}", examples=["i-node01"], description="Node id."
    )
    user: Annotated[
        int,
        UserID(default_uid=_file_perm.st_uid),
        Field(default=_file_perm.st_uid, description="Process owner."),
    ]
    group: Annotated[
        int,
        GroupID(default_gid=_file_perm.st_gid),
        Field(default=_file_perm.st_uid, description="Process group."),
    ]

    bind_host: str = Field(
        default="0.0.0.0",
        description="Bind address of the port opened on behalf of wsproxy worker",
    )
    advertised_host: str = Field(
        examples=["example.com"],
        description="Hostname to be advertised to client",
    )

    bind_api_port: int = Field(
        default=5050,
        description="Port number to bind for API server",
    )
    internal_api_port: int = Field(
        default=15050, description="Port number to bind for internal API server"
    )
    advertised_api_port: Optional[int] = Field(
        default=None,
        examples=[15050],
        description="API port number reachable from client",
    )

    bind_proxy_port_range: tuple[int, int] = Field(
        default=(10200, 10300),
        description="Port number to bind for actual traffic",
    )
    advertised_proxy_port_range: Optional[tuple[int, int]] = Field(
        default=None,
        examples=[[20200, 20300]],
        description="Traffic port range reachable from client",
    )

    protocol: ProxyProtocol = Field(default=ProxyProtocol.HTTP, description="Proxy protocol")

    jwt_encrypt_key: str = Field(
        examples=["50M3G00DL00KING53CR3T"],
        description="JWT encryption key",
    )
    permit_hash_key: str = Field(
        examples=["50M3G00DL00KING53CR3T"],
        description="Permit hash key",
    )

    api_secret: str = Field(
        examples=["50M3G00DL00KING53CR3T"],
        description="API secret",
    )
    aiomonitor_termui_port: int = Field(
        gt=0,
        lt=65536,
        description="Port number for aiomonitor termui server.",
        default=38500,
    )
    aiomonitor_webui_port: int = Field(
        gt=0,
        lt=65536,
        description="Port number for aiomonitor webui server.",
        default=39500,
    )

    @property
    def advertised_api_addr(self) -> str:
        """
        Returns the advertised API address.
        """
        port = self.advertised_api_port or self.bind_api_port
        return f"http://{self.advertised_host}:{port}"


class PyroscopeConfig(BaseConfigModel):
    enabled: bool = Field(default=False, description="Enable pyroscope profiler.")
    app_name: Optional[str] = Field(default=None, description="Pyroscope app name.")
    server_addr: Optional[str] = Field(default=None, description="Pyroscope server address.")
    sample_rate: Optional[int] = Field(default=None, description="Pyroscope sample rate.")


class ServerConfig(BaseConfigModel):
    wsproxy: Annotated[WSProxyConfig, Field(default_factory=WSProxyConfig)]
    pyroscope: PyroscopeConfig = Field(default_factory=PyroscopeConfig)
    logging: LoggingConfig = Field(default_factory=LoggingConfig)
    debug: DebugConfig = Field(default_factory=DebugConfig)


def load(config_path: Optional[Path] = None, log_level: LogLevel = LogLevel.NOTSET) -> ServerConfig:
    # Determine where to read configuration.
    raw_cfg, _ = config.read_from_file(config_path, "wsproxy")

    config.override_key(raw_cfg, ("debug", "enabled"), log_level == LogLevel.DEBUG)
    if log_level != LogLevel.NOTSET:
        config.override_key(raw_cfg, ("logging", "level"), log_level)
        config.override_key(raw_cfg, ("logging", "pkg-ns", "ai.backend"), log_level)

    # Validate and fill configurations
    # (allow_extra will make configs to be forward-copmatible)
    try:
        cfg = ServerConfig(**raw_cfg)
        if cfg.debug.enabled:
            print("== WSProxy configuration ==", file=sys.stderr)
            print(pformat(cfg.model_dump()), file=sys.stderr)
    except ValidationError as e:
        print(
            "ConfigurationError: Could not read or validate the wsproxy local config:",
            file=sys.stderr,
        )
        print(pformat(e.errors()), file=sys.stderr)
        raise click.Abort()
    else:
        return cfg


class Undefined:
    pass


class UnsupportedTypeError(RuntimeError):
    pass


def generate_example_json(
    schema: type[BaseConfigModel] | types.GenericAlias | types.UnionType, parent: list[str] = []
) -> dict | list:
    if isinstance(schema, types.UnionType):
        return generate_example_json(typing.get_args(schema)[0], parent=[*parent])
    elif isinstance(schema, types.GenericAlias):
        if typing.get_origin(schema) is not list:
            raise RuntimeError("GenericAlias other than list not supported!")
        return [generate_example_json(typing.get_args(schema)[0], parent=[*parent])]
    elif issubclass(schema, BaseConfigModel):
        res = {}
        for name, info in schema.model_fields.items():
            config_key = [*parent, name]
            assert info.annotation
            alternative_example = Undefined
            if info.examples:
                res[name] = info.examples[0]
            elif info.default != PydanticUndefined:
                alternative_example = info.default
            if name not in res:
                try:
                    res[name] = generate_example_json(info.annotation, parent=config_key)
                except RuntimeError:
                    if alternative_example != Undefined:
                        res[name] = alternative_example
                    else:
                        raise
        return res
    else:
        raise UnsupportedTypeError(str(schema))
